__all__ = ["run"]

import argparse
import os
import pickle
import random
import subprocess
import sys
import time

sys.path.insert(0, os.path.join(os.path.expanduser("~"), ".sumsjob"))
import config
from .sinfo import gpu_available
from .utils import local_cmdline


def get_machine(machine, gpuid, verbose=0):
    if machine is None:
        machine, gpuid = gpu_available(first_only=True, verbose=verbose)
    else:
        if gpuid == "-1":
            if verbose > 0:
                print("Server: %s, CPU" % machine)
            gpuid = ""
        else:
            if verbose > 0:
                print("Server: %s, gpuid: %s" % (machine, gpuid))
    return machine, gpuid


def push_files(src, machine, dest, options="", verbose=0):
    if verbose == 2:
        options = "-avhP " + options
    else:
        options = "-a " + options
    if config.LAN is not None:
        cmd = "rsync {} {} {}:{}".format(options, src, config.LAN, dest)
    else:
        cmd = "rsync {} {} {}:{}".format(options, src, machine, dest)
    if verbose == 2:
        print(cmd)
    subprocess.check_call(cmd, shell=True)


def pull_files(machine, runpath, verbose=0):
    if verbose == 2:
        options = "-avhP " + config.files_pull
    else:
        options = "-a " + config.files_pull
    if config.LAN is not None:
        cmd = "rsync {} {}:{}/ .".format(options, config.LAN, runpath)
    else:
        cmd = "rsync {} {}:{}/ .".format(options, machine, runpath)
    if verbose == 2:
        print(cmd)
    subprocess.check_call(cmd, shell=True)


def run_one(
    jobpy,
    jobname=None,
    machine=None,
    gpuid=None,
    interact=True,
    inobj=None,
    infile="in.pickle",
    verbose=0,
):
    machine, gpuid = get_machine(machine, gpuid, verbose=verbose)
    if machine is None:
        if verbose > 0:
            print("Failed to submit.")
            sys.stdout.flush()
        return

    if jobname is None:
        jobname = "%d%04x" % (time.time(), random.randint(0, int("ffff", 16)))
    runpath = os.path.join(config.path_prefix, jobname)

    push_files("./", machine, runpath, options=config.files_push, verbose=verbose)

    if inobj is not None:
        tmpfile = "/tmp/sumsjob." + jobname
        with open(tmpfile, "wb") as f:
            pickle.dump(inobj, f)
        push_files(tmpfile, machine, os.path.join(runpath, infile), verbose=verbose)
        subprocess.check_call(["rm", "-f", tmpfile])
    sys.stdout.flush()

    if interact:
        cmd = "cd {} && CUDA_VISIBLE_DEVICES={} {} {} 2>&1 | tee {}.log".format(
            runpath, gpuid, config.cmd, jobpy, jobname
        )
        cmd = local_cmdline(machine, cmd, interact=True, verbose=verbose)
        subprocess.check_call(cmd, shell=True)
        pull_files(machine, runpath, verbose=verbose)
    else:
        # Start a new screen session in the backgroun; run the code;
        # As soon as the code finishes, the screen session terminates.
        # - https://serverfault.com/questions/104668/create-screen-and-run-command-without-attaching
        # - https://askubuntu.com/questions/62562/run-a-program-with-gnu-screen-and-immediately-detach-after
        cmd = "cd {} && CUDA_VISIBLE_DEVICES={} {} {} 2>&1 | tee {}.log".format(
            runpath, gpuid, config.cmd, jobpy, jobname
        )
        cmd = f'screen -dmS sumsjob-{jobname} bash -c "{cmd}"'
        cmd = local_cmdline(machine, cmd, verbose=verbose)
        subprocess.check_call(cmd, shell=True)

        # If we want to keep the screen session, send command to a detached screen session
        # - https://raymii.org/s/snippets/Sending_commands_or_input_to_a_screen_session.html
        # - https://askubuntu.com/questions/983063/start-a-screen-session-and-run-a-script-without-attaching-to-it
        # Start a detached screen
        # cmd_server = f"screen -dmS {jobname}"
        # cmd = f"ssh {machine} '{cmd_server}'"
        # if verbose == 2:
        #     print(cmd)
        # subprocess.check_call(cmd, shell=True)
        # Send command to that session session
        # cmd_run = "cd {} && CUDA_VISIBLE_DEVICES={} {} {} 2>&1 | tee {}.log".format(
        #     runpath, gpuid, config.cmd, jobpy, jobname
        # )
        # cmd_server = f'screen -S {jobname} -X stuff "{cmd_run}^M"'
        # cmd = f"ssh {machine} '{cmd_server}'"
        # if verbose == 2:
        #     print(cmd)
        # subprocess.check_call(cmd, shell=True)

        # Show sessions
        cmd = "screen -list"
        cmd = local_cmdline(machine, cmd, verbose=verbose)
        subprocess.check_call(cmd, shell=True)
        print(f"Server: {machine}")
        print(f"Job: {jobname}")
    return jobname


def run(
    jobpy,
    jobname=None,
    machine=None,
    gpuid=None,
    interact=True,
    inobj=None,
    infile="in.pickle",
    retry_num=1,
    retry_period=60,
    verbose=0,
):
    for i in range(retry_num):
        jobname = run_one(
            jobpy,
            jobname=jobname,
            machine=machine,
            gpuid=gpuid,
            interact=interact,
            inobj=inobj,
            infile=infile,
            verbose=verbose,
        )
        # Successed
        if jobname:
            return jobname
        # Failed
        if i < retry_num - 1:
            print("Wait for %d s...\n" % retry_period)
            time.sleep(retry_period)


def main():
    parser = argparse.ArgumentParser(
        description="Submit a job to GPU servers for execution."
    )
    parser.add_argument("jobfile", help="File to be run")
    parser.add_argument(
        "jobname",
        nargs="?",
        help="Job name, and also the folder name of the job. If not provided, a random number will be used.",
    )
    parser.add_argument(
        "-i", "--interact", action="store_true", help="Run the job in interactive mode"
    )
    parser.add_argument("-s", "--server", help="Server host name")
    parser.add_argument("--gpuid", help="GPU ID to be used; -1 to use CPU only")
    parser.add_argument(
        "-n",
        "--num_retry",
        default=1000,
        type=int,
        help="Number of times to retry the submission (Default: 1000)",
    )
    parser.add_argument(
        "-p",
        "--period_retry",
        default=600,
        type=int,
        help="Waiting time (seconds) between two retries after each retry failure (Default: 600)",
    )
    args = parser.parse_args()

    run(
        args.jobfile,
        jobname=args.jobname,
        machine=args.server,
        gpuid=args.gpuid,
        interact=args.interact,
        retry_num=args.num_retry,
        retry_period=args.period_retry,
        verbose=2,
    )


if __name__ == "__main__":
    main()



__all__ = ["sacct"]

import datetime
import os
import subprocess
import sys

sys.path.insert(0, os.path.join(os.path.expanduser("~"), ".sumsjob"))
import config
from .utils import local_cmdline


def order_by_start(jobs):
    for job in jobs:
        job["Start datetime"] = datetime.datetime.strptime(
            job["Start"], "%m/%d/%Y %I:%M:%S %p"
        )
    return sorted(jobs, key=lambda e: e["Start datetime"])


def sacct(verbose=0):
    jobs = []
    for m in config.servers:
        cmd = "screen -list"
        cmd = local_cmdline(m, cmd, verbose=0)
        process = subprocess.run(cmd, capture_output=True, shell=True, text=True)

        # Example:
        # There is a screen on:
        #         47065.sumsjob-jobname   (12/27/2021 02:02:41 AM)        (Detached)
        # 1 Sockets in /run/screen/S-lu.
        # Or
        # No Sockets found in /run/screen/S-lu.
        lines = process.stdout.strip().split("\n")
        lines = filter(lambda l: "sumsjob-" in l, lines)
        if not lines:
            continue

        lines = map(lambda l: l.strip(), lines)
        for l in lines:
            session_name, creation_time = l.split("\t")[:2]
            job_name = session_name.split("sumsjob-", 1)[1]
            creation_time = creation_time[1:-1]
            jobs.append({"Server": m, "JobName": job_name, "Start": creation_time})

    jobs = order_by_start(jobs)
    if verbose > 0:
        print("Server   JobName          Start")
        print("-------- ---------------- ----------------------")
        for job in jobs:
            job_name = job["JobName"]
            if len(job_name) > 16:
                job_name = job_name[:15] + "+"
            print(f"{job['Server']:<8} {job_name:<16} {job['Start']}")
    return jobs


def main():
    sacct(verbose=1)


if __name__ == "__main__":
    main()



__al__ = ["scancel"]

import argparse
import subprocess

from .sacct import sacct
from .utils import local_cmdline


def scancel(jobname):
    jobs = sacct()
    for job in jobs:
        if job["JobName"] == jobname:
            cmd = f"screen -S sumsjob-{jobname} -X quit"
            cmd = local_cmdline(job["Server"], cmd, verbose=0)
            subprocess.run(cmd, shell=True, check=True)
            print(f"Job {jobname} on {job['Server']} cancelled.")
            return
    print(f"Job {jobname} not found.")


def main():
    parser = argparse.ArgumentParser(description="Cancel a running job.")
    parser.add_argument("jobname", help="Job name")
    args = parser.parse_args()

    scancel(args.jobname)


if __name__ == "__main__":
    main()



__all__ = ["gpu_available"]

import os
import subprocess
import sys

sys.path.insert(0, os.path.join(os.path.expanduser("~"), ".sumsjob"))
import config
from .utils import local_cmdline


def exclude_gpus(lines):
    if config.gpus_exclude:
        lines = [
            l for l in lines if all(map(lambda gpu: gpu not in l, config.gpus_exclude))
        ]
    return lines


def gpustat(machine, stat):
    # Example of stat:
    # chitu                       Fri Dec 31 01:33:24 2021  470.74
    # [0] NVIDIA GeForce RTX 3080 | 70'C,  67 % |  3637 / 10018 MB | shuaim:python3/3589(689M)
    # [1] NVIDIA GeForce RTX 3080 | 66'C,  37 % |  6412 / 10014 MB | shuaim:python3/3589(361M)
    lines = stat.strip().split("\n")
    lines = exclude_gpus(lines)
    avail = None
    for l in lines[1:]:
        contents = l.split("|")
        utilization = int(contents[1].split(",", 1)[1].split("%")[0]) / 100
        memory = contents[2].split("/", 1)
        memory_used = int(memory[0])
        memory_total = int(memory[1].split("MB")[0])
        if (
            1 - utilization >= config.gpu_utilization
            and memory_total - memory_used >= config.gpu_memory * 1024
        ):
            avail = int(l[1])
    return machine, "\n".join(lines), avail


def gpuresource():
    for m in config.servers:
        cmd = "gpustat -cup"
        cmd = local_cmdline(m, cmd, verbose=0)
        try:
            stat = subprocess.check_output(cmd, shell=True)
        except subprocess.CalledProcessError:
            print("Failed to get the GPU status on {}\n".format(m))
            continue
        if sys.version_info >= (3, 0):
            stat = stat.decode("utf-8")
        yield gpustat(m, stat)


def gpu_available(first_only=False, verbose=0):
    gm, gavail = None, None
    for m, stat, avail in gpuresource():
        if verbose == 2:
            print(stat)
            print("")
        if gavail is None and avail is not None:
            gm, gavail = m, avail
            if first_only:
                break
    if gm is None:
        if verbose > 0:
            print("No available GPU.")
        return None, None
    else:
        if verbose > 0:
            print("Available GPU: {} [{}]\n".format(gm, gavail))
        return gm, gavail


def main():
    gpu_available(verbose=2)


if __name__ == "__main__":
    main()



import os
import sys

sys.path.insert(0, os.path.join(os.path.expanduser("~"), ".sumsjob"))
import config


def local_cmdline(machine, cmd_server, interact=False, verbose=0):
    if interact:
        cmd = f"ssh -tX {machine} '{cmd_server}'"
    else:
        cmd = f"ssh {machine} '{cmd_server}'"
    if config.LAN is not None:
        if interact:
            cmd = 'ssh -tX {} "{}"'.format(config.LAN, cmd)
        else:
            cmd = 'ssh {} "{}"'.format(config.LAN, cmd)
    if verbose == 2:
        print(cmd)
    return cmd



# Host names of servers
servers = ["server1", "server2"]

# In some situations, you may need to first log in a local area network (LAN) to access
# the servers. In this case, provide the host name of LAN. Also, the same home folder
# should be shared between LAN and all servers.
# None if LAN is not required.
LAN = None

# GPUs to be excluded
# e.g., gpus_exclude = ["Quadro"] to exclude all GPUs whose name contains "Quadro"
# Emply list if no GPUs to be excluded
gpus_exclude = []

# Criterion of using a GPU:
# (1) its utilization available > gpu_utilization
# (2) its memory available (GB) > gpu_memory
# By default, the GPU utilization should <= 10%, i.e., utilization available >= 90%
gpu_utilization = 0.90
# By default, the GPU should have >= 4 GB memory available
gpu_memory = 4

# Root folder of the files
# A new folder will be created in this root folder to store your codes.
# For example, if the jobname is "myjob", then the folder is "~/scratch/myjob".
path_prefix = "~/scratch"

# Command to run your job file, i.e., the python path
cmd = "python"

# Options for [rsync](https://linux.die.net/man/1/rsync) to control what files will be
# copied from your computer to the server, and what files will be transferred back from
# the server to your computer.
# Files copied from your computer to the server, i.e., files required to run the job
# By default, all .py files in the current folder, including .py files in the subfolders.
files_push = "--include='[^.]*/' --include='*.py' --exclude='*'"
# Files transferred back from the server to your computer, i.e., new files you need
# By default, all .dat and .log files
files_pull = "--include='*.dat' --include='*.log' --exclude='*'"



import io
import os
from setuptools import setup
from setuptools import find_packages


def read(rel_path: str) -> str:
    here = os.path.abspath(os.path.dirname(__file__))
    # intentionally *not* adding an encoding option to open, See:
    #   https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    with open(os.path.join(here, rel_path)) as fp:
        return fp.read()


def get_version(rel_path: str) -> str:
    for line in read(rel_path).splitlines():
        if line.startswith("__version__"):
            # __version__ = "0.11.2"
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    raise RuntimeError("Unable to find version string.")


version = get_version("sumsjob/__about__.py")

with io.open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="SumsJob",
    version=version,
    description="A simple Linux command-line utility which submits a job to one of the multiple GPU servers",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Lu Lu",
    author_email="lululxvi@gmail.com",
    url="https://github.com/lululxvi/sumsjob",
    download_url="https://github.com/lululxvi/deepxde/tarball/v" + version,
    license="GPL-3.0",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: End Users/Desktop",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Topic :: Utilities",
    ],
    keywords=[
        "Command-line utility",
        "Multiple servers",
        "GPU",
        "Job submission",
    ],
    packages=find_packages(),
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "sacct=sumsjob.sacct:main",
            "scancel=sumsjob.scancel:main",
            "sinfo=sumsjob.sinfo:main",
            "srun=sumsjob.srun:main",
        ]
    },
)
